
/**********************************************************************

Copyright (C) 1995, 1996  E. O. Lawrence Berkeley National Laboratory.  
All Rights Reserved.  Permission to copy and modify this software and 
its documentation (if any) is hereby granted, provided that this notice
is retained thereon and on all copies.  

This software is provided as a professional academic contribution for 
joint exchange.   Thus it is experimental and scientific in nature, 
undergoing development, and is provided "as is" with no warranties of 
any kind whatsoever, no support, promise of updates or printed 
documentation.

This work is supported by the U. S. Department of Energy under contract 
number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
University of California.


	Author: Wes Bethel
		Lawrence Berkeley National Laboratory
                Berkeley, California 

  "this software is 100% hand-crafted by a human being in the USA"

***********************************************************************/

#include <stdio.h>
#include <limits.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>

#include "fastrak.h"
#include "matrix.h"

#define M_PI 3.14159
#define DEGREES_TO_RADIANS(a) ((a)*M_PI/180.)
#define A_SHITLOAD INT_MAX
#define NLOOPS 1000
#define MAX(a,b) ((a) > (b) ? (a) : (b))

static int do_reset = 1;
static int do_filter = 0;
static int do_onthefly_realignment = 0;
static int station_to_print = 1;
static int do_3dmouse=0;
static int do_print = 1;

static matrix4x4 mlast[4],mthis[4],minverse[4],mwork[4];


main(int ac,
     char *av[])
{
    int status,i,nstations,j,k;
    station_info sa[MAX_STATIONS];
    int active_stations[MAX_STATIONS];
    int bmask[8];
    time_t start,finish;
    struct timeval tv1,tv2;
    float delta_azimuth[4], delta_elevation[4], delta_roll[4];

    parse_cmd_line(ac,av);

    /* open the device */
    status = open_fastrak();

    if (do_3dmouse)
	status = open_3dmouse();

    /* reset the device */
    if (do_reset)
    {
        status = reset_fastrak();
	if (status == 0)
	  exit(0);
    }


    /* configure the device */
    memset((char *)sa,0,sizeof(station_info)*MAX_STATIONS);
    nstations = configure_fastrak_stations(sa);

    fprintf(stderr,"\tThere are %d stations reported active.\n",nstations);

    /* enable filtering, if requested. */
    if (do_filter)
    {
	filter_enable();
    }

    /* do sanity checking on station_to_print, convert from 1..n to
     an index in the stations array. */

    for (i=0;i<nstations;i++)
	if (station_to_print == sa[i].station_num)
	    break;
    if (i == nstations)
    {
	fprintf(stderr," there is no station number %d to print. will revert to station 1. \n",station_to_print);
	station_to_print = 0;	/* the assumption being that sa[0] == station 1 */
    }
    else
	station_to_print = i;

    start = time(NULL);
    /* do some stuff */
    for (i=0;i<NLOOPS;i++)
    {
	/* do the mouse events */
	if (do_3dmouse)
	{
	    memset((char *)bmask,0,sizeof(int)*8);
	    status = get_mouse_data(bmask);
	    if (status)
	    {
		if (bmask[0])
		    fprintf(stderr,"LEFT \n");
		if (bmask[1])
		    fprintf(stderr,"MIDDLE \n");
		if (bmask[2])
		    fprintf(stderr,"RIGHT \n");
		if (bmask[3])
		    fprintf(stderr,"TOP \n");
		if (bmask[4])
		    fprintf(stderr,"BOTTOM \n");
		fflush(stderr);
	    }
	}

	if (gettimeofday(&tv1) != 0)
	    fprintf(stderr," couldn't get time value \n");
	
	status = get_fastrak_data(sa);
	
	if (gettimeofday(&tv2) != 0)
	    fprintf(stderr," couldn't get time value \n");
#if 0
	fprintf(stderr,"\nelapsed time in transfer: %d\t\t%d\n",tv2.tv_sec-tv1.tv_sec,tv2.tv_usec-tv1.tv_usec);

#endif
	for (j=0;j<3;j++)
	    if (sa[0].rot[j] < 0.)
		sa[0].rot[j] += 360.;
	
/*		fprintf(stderr," using %g %g %g \n",sa[0].rot[0],sa[0].rot[1],sa[0].rot[2]); */

	if (i == 0)		/* assume status==1 */
	{
	    /* build initial xformation matrix */
	    for (j=0;j<nstations;j++)
	    {
		build_composite_matrix(mlast+j,sa[j].rot);
	    }
	    continue;
	}


	if (status)
	{
	    for (j=0;j<nstations;j++)
	    {
		build_composite_matrix(mthis+j,sa[j].rot);

		recompute_base_rotations(mthis+j,mlast+j,delta_azimuth+j,
					 delta_elevation+j,delta_roll+j);
	    }
	    memcpy((char *)mlast,(char *)mthis,sizeof(matrix4x4)*4);
	    
/*	    fprintf(stderr,"\t%6.3f\t%6.3f\t%6.3f\t%06.3f\t%06.3f\t%06.3f\r",delta_azimuth[station_to_print],delta_elevation[station_to_print],delta_roll[station_to_print],sa[station_to_print].rot[0],sa[station_to_print].rot[1],sa[station_to_print].rot[2]);

	    fflush(stderr); */
	}
	else
	    fprintf(stderr,"\t failed to grab a packet!!! \n");
    }
    
    finish = time(NULL);

    fprintf(stderr,"approx %g events per second.\n",(float)((float)i/(float)(finish-start)));

    /* close the device */
    status = close_fastrak();
}

int
parse_cmd_line(int ac, char *av[])
{
    int count,i;

    i = 1;
    for (i=1;i<ac;)
    {
       if (strncmp(av[i],"-nr",3) == 0)
       {
	   do_reset = 0;
	   fprintf(stderr," turning off the fastrak reset.\n");
	   i++;
       }
       else if (strncmp(av[i],"-f",2) == 0)
       {
	   do_filter = 1;
	   fprintf(stderr," positional and rotational filtering enabled. \n");
	   i++;
       }
       else if (strncmp(av[i],"-a",2) == 0)
       {
	   fprintf(stderr," doing on-the-fly alignment reference frame reset (relative rotational values will be generated.\n");
	   do_onthefly_realignment = 1;
	   i++;
       }
       else if (strncmp(av[i],"-s",2) == 0)
       {
	   i++;
	   sscanf(av[i],"%d",&station_to_print);
	   fprintf(stderr," will print info from station %d.\n",station_to_print);
	   i++;
       }
       else if (strncmp(av[i],"-m",2) == 0)
       {
	   fprintf(stderr," will read button events from the 3d mouse.\n");
	   do_3dmouse = 1;
	   i++;
       }
       else if (strncmp(av[i],"-noprint",strlen("-noprint")) == 0)
       {
	   fprintf(stderr," disabling the printing of tracker events.\n");
	   do_print = 0;
	   i++;
       }
       else
       {
	   fprintf(stderr," unrecognized command line option<%s> \n",av[i]);
	   i++;
       }
    }
}

int
print_station_info(station_info *sa)
{
    fprintf(stderr,"\n\tStation %d reports:\n",sa->station_num);
    if (sa->flags & POSITION_FLAG)
	fprintf(stderr,"\t\tPos(x,y,z):\t%g\t%g\t%g\n",sa->pos[0],sa->pos[1],sa->pos[2]);
    if (sa->flags & ROTATION_FLAG)
	fprintf(stderr,"\t\tRot(x,y,z):\t%g\t%g\t%g\n",sa->rot[0],sa->rot[1],sa->rot[2]);
    if (sa->flags & XDIRCOS_FLAG)
	fprintf(stderr,"\t\tXDIR(x,y,z):\t%g\t%g\t%g\n",sa->xdircos[0],sa->xdircos[1],sa->xdircos[2]);
    if (sa->flags & YDIRCOS_FLAG)
	fprintf(stderr,"\t\tYDIR(x,y,z):\t%g\t%g\t%g\n",sa->ydircos[0],sa->ydircos[1],sa->ydircos[2]);
    if (sa->flags & ZDIRCOS_FLAG)
	fprintf(stderr,"\t\tZDIR(x,y,z):\t%g\t%g\t%g\n",sa->zdircos[0],sa->zdircos[1],sa->zdircos[2]);
}
#if 0
int
build_composite_matrix(matrix4x4 *m,float *r)
{
    /**
      * NOTES:
      *
      * what we're going to do here is build a transformaion matrix and
      * put it into "m".  this transformation matrix is formed by the absolute
      * rotation angles present in float r[3].  the way in which r[] is
      * interpreted is r[0] = azimuth, r[1] = elevation, r[2] = roll.
      * furthermore, we're making some assumptions with regard to the
      * fastrak device.  these are:
      *
      * 1. the alignment reference frame for all stations is the power on
      *    reference frame: a right handed coordinate systems with +z
      *    pointing up.
      *
      * 2. azimuth represents angular changes about the z axis of the
      *    alignment reference frame; elevation represents changes aboiut
      *    the y axis of the alignment reference frame; roll represents
      *    angular changes about the x-axis of the alignment reference
      *    frame.
      * 3. the angular values presented in r[] are taken in the order of
      *    azimuth, elevation then roll.  the transformation matrix we
      *    construct below will be computed in this order.
      *
      * the sequence of computations below is valid only if all these
      * assumptions are TRUE.
      *
      * the routine which is called subsequent to this routine,
      * "recompute_base_rotations" will recompute delta azimuth, elevation
      * and roll values making use of the same 3 fastrak assumptions as this
      * routine.
    **/
      
    matrix4x4 azimuth, elevation, roll;
    double t,c,s;
    extern double cos(),sin();

    /* do azimuth - according to the fastrak manual, azimuth angles
     occur about the "z" axis.*/
    identity_4x4(&azimuth);
    t = DEGREES_TO_RADIANS(r[0]);
    c = cos(t);
    s = sin(t);
    azimuth.m[0][0] = azimuth.m[1][1] = c;
    azimuth.m[0][1] = s;
    azimuth.m[1][0] = -s;

    /* do elevation */
    t = DEGREES_TO_RADIANS(r[1]); /* rotation about y axis */
    c = cos(t);
    s = sin(t);
    identity_4x4(&elevation);
    elevation.m[0][0] = elevation.m[2][2] = c;
    elevation.m[0][2] = -s;
    elevation.m[2][0] = s;

    /* do roll - according to the manual, roll occurs about the X axis. */
    identity_4x4(&roll);
    t = DEGREES_TO_RADIANS(r[2]); /* rotation about x axis */
    c = cos(t);
    s = sin(t);
    roll.m[1][1] = roll.m[2][2] = c;
    roll.m[1][2] = s;
    roll.m[2][1] = -s;

    identity_4x4(m);
#if 0
    mmul_4x4(m,&azimuth);
    mmul_4x4(m,&elevation);
    mmul_4x4(m,&roll);
#endif
    mmul_4x4(m,&roll);
    mmul_4x4(m,&elevation);
    mmul_4x4(m,&azimuth);
}

int
recompute_base_rotations(mthis,mlast,azimuth,elevation,roll)
matrix4x4 *mthis;
matrix4x4 *mlast;
float *azimuth;
float *elevation;
float *roll;
{
    /**
     * "roll" occurs about the X axis
     * "azimuth" occurs about the Z axis
     * "elevation" occurs about the Y axis
     **/
    
    extern double acos();
    int k;
    double t1,t2;
    vector3 xold,yold,zold,xnew,ynew,znew;
    
    for (k=0;k<3;k++)
    {
	xold.v[k] = mlast->m[k][0];
	yold.v[k] = mlast->m[k][1];
	zold.v[k] = mlast->m[k][2];
		    
	xnew.v[k] = mthis->m[k][0];
	ynew.v[k] = mthis->m[k][1];
	znew.v[k] = mthis->m[k][2];
    }

    /* compute amount of change about the Z axis - this is azimuth */
    vec3dot(&xold,&xnew,&t1);
    vec3dot(&yold,&ynew,&t2);
    t1 = MAX(t1,t2);
    if (t1 > 1.)
	t1 = 1.;
    t1 = acos(t1);
    vec3dot(&xnew,&yold,&t2);
    t1 *= ((t2 > 0.) ? 1. : -1.);
    *azimuth = (float)t1;

    /* compute the amount of change about the Y axis. this is elevation */
    vec3dot(&xold,&xnew,&t1);
    vec3dot(&zold,&znew,&t2);
    t1 = MAX(t1,t2);
    if (t1 > 1.)
	t1 = 1.;
    t1 = acos(t1);
    vec3dot(&xnew,&zold,&t2);
    t1 *= ((t2 > 0.) ? 1. : -1.);
    *elevation = (float)t1;

    /* compute the amount of change about the X axis.  this is roll. */
    vec3dot(&yold,&ynew,&t1);
    vec3dot(&zold,&znew,&t2);
    t1 = MAX(t1,t2);
    if (t1 > 1.)
	t1 = 1.;
    t1 = acos(t1);
    vec3dot(&znew,&yold,&t2);
    t1 *= ((t2 > 0.) ? 1. : -1.);
    *roll = (float)t1;
}
#endif
